package org.processmining.lpengines.interfaces;

import java.util.Map;

import org.processmining.framework.util.Pair;

/**
 * An LPEngine allows for solving LP-problems. The engineType has no knowledge
 * of any form of "variable/constraints mapping".
 */
public interface LPEngine extends Cloneable {

	/**
	 * Engine implemented by implementing class.
	 */
	public enum EngineType {
		LPSOLVE("LPSolve", 5.5d), LPSOLVE_DISPOSABLE("LPSolve (Disposable)", 5.5d);

		private final String name;
		private final double version;

		EngineType(String name, double version) {
			this.name = name;
			this.version = version;
		}

		public String givenName() {
			return this.name;
		}

		@Override
		public String toString() {
			return givenName() + ", version " + this.version;
		}
	}

	/**
	 * Possible types of objectives of an LP
	 */
	public enum ObjectiveTargetType {
		MAX, MIN
	}

	/**
	 * Enumeration of operators to use in objectives function and constraints.
	 */
	public enum Operator {
		EQUAL, GREATER_EQUAL, LESS_EQUAL
	}

	/**
	 * Possible variable types within an LP
	 */
	public enum VariableType {
		BOOLEAN, INTEGER, REAL
	}

	void accept(LPEngineVisitor visitor);

	/**
	 * Add a constraints to the LP Engine
	 *
	 * @param constraint
	 *            to add
	 * @param operator
	 *            of constraints
	 * @param value
	 *            target value of constraints
	 * @return row index of the constraints
	 */
	public int addConstraint(double[] constraint, Operator operator, double value);

	/**
	 * Add a constraints to the LP Engine
	 *
	 * @param constraint
	 *            map of variable to double value
	 * @param operator
	 *            of constraints
	 * @param value
	 *            target value of constraints
	 * @return row index of the constraints
	 */
	public int addConstraint(Map<Integer, Double> constraint, Operator operator, double value);

	/**
	 * Add a variable. Note: the double array should equal the number of rows in
	 * the LP. The type parameter specifies the variable's type
	 *
	 * @param column
	 *            an array of double's specifying the column
	 * @param type
	 *            type of the variable
	 * @return column index of variable, -1 upon failure
	 */
	public int addVariable(double[] column, VariableType type);

	/**
	 * Add a variable. The map, maps a variable index to a coefficient. The type
	 * parameter specifies the variable's type
	 *
	 * @param nonZeroValues
	 *            mapping of row number to non-zero elements
	 * @return column index of lp variable, -1 upon failure.
	 */
	public int addVariable(Map<Integer, Double> nonZeroValues, VariableType type);

	public LPEngine clone() throws CloneNotSupportedException;

	/**
	 * destroy the ILP problem and (try to) free the memory;
	 */
	void destroy();

	/**
	 * Return an empty constraint ([0,0,...0]) with the appropriate length for
	 * addition.
	 *
	 * @return constraint
	 */
	public double[] emptyConstraint();

	/**
	 * Return an empty objective array ([0,0,...0]) with the appropriate length
	 * for addition.
	 *
	 * @return objective array
	 */
	public double[] emptyVariable();

	/**
	 * Returns what engineType is actually used by the implementing class. This
	 * knowledge can be used for variable mappings etc.
	 *
	 * @return the engineType used.
	 */
	public EngineType engineType();

	public double getVariableLowerBound(int variableIndex);

	public double getVariableUpperBound(int variableIndex);

	/**
	 * Assess whether is at least one solution
	 *
	 * @return true iff the LP is feasible
	 */
	public boolean isFeasible();

	/**
	 * Given a specific engine type (e.g. LPSOLVE, GUROBI, etc.) this utility
	 * returns the minimal constraint index. LPSOLVE for example can not have a
	 * variable with index 0.
	 *
	 * @return minimal constraint index
	 */
	public int minimalConstraintIndex();

	/**
	 * Given a specific engine type (e.g. LPSOLVE, GUROBI, etc.) this utility
	 * returns the minimal variable index. LPSOLVE for example can not have a
	 * variable with index 0.
	 *
	 * @return minimal variable index
	 */
	public int minimalVariableIndex();

	/**
	 * Query the total number of constraints (== rows in matrix)
	 *
	 * @return number of constraints
	 */
	public int numConstraints();

	/**
	 * Query the total number of variables (== columns in matrix)
	 *
	 * @return number of variables
	 */
	public int numVariables();

	/**
	 * pop a constraint out of the engine (that is the constraint with the max.
	 * constr index)
	 */
	public void pop();

	/**
	 * Print the current state of the LP.
	 */
	public void print();

	/**
	 * Remove the given constraints. Be careful to update your constraints
	 * administration!
	 *
	 * @param row
	 *            to be removed
	 * @return success indication
	 */
	public boolean removeConstraint(int row);

	/**
	 * Remove a variable from the LP. Note that this causes a shift in all
	 * variables (columns) with an id > column!
	 *
	 * @param column
	 *            to remove
	 * @return success indication
	 */
	public boolean removeVariable(int column);

	/**
	 * Set the given row with the provided constraints Note: the size of the
	 * array should equal the number of constraints.
	 *
	 * @param row
	 *            to set
	 * @param constraint
	 *            new row
	 * @return success indication
	 */
	public boolean setConstraint(int row, double[] constraint);

	/**
	 * Set the given row with the provided constraint and rhs.
	 * 
	 * @param row
	 *            to set
	 * @param constraint
	 *            new constraint
	 * @param rhs
	 *            right hand side
	 * @return
	 */
	public boolean setConstraint(int row, double[] constraint, double rhs);

	/**
	 * Set the given row with the provided constraint, operator and rhs
	 * 
	 * @param row
	 *            to set
	 * @param constraint
	 *            new constraint
	 * @param op
	 *            operator
	 * @param rhs
	 *            right hand side
	 * @return
	 */
	public boolean setConstraint(int row, double[] constraint, Operator op, double rhs);

	/**
	 * Renew the given row with the provided constraints. Note: the size of the
	 * map should equal the number of constraints.
	 *
	 * @param row
	 *            number of constraints to renew
	 * @param constraint
	 *            to update
	 * @return success indication
	 */
	public boolean setConstraint(int row, Map<Integer, Double> constraint);

	/**
	 * Set the given constraint's operator (e.g. > < ==)
	 *
	 * @param row
	 *            to set
	 * @param op
	 *            operator
	 * @return success indication
	 */
	public boolean setConstraintOperator(int row, Operator op);

	/**
	 * Set a variable's lower-bound.
	 *
	 * @param column
	 *            identifier of variable
	 * @param lowerBound
	 *            to assing to variable
	 * @return engineType manipulation succeeded
	 */
	public boolean setLowerBound(int column, double lowerBound);

	/**
	 * Set a cell within the engine.
	 *
	 * @param rowIndex
	 * @param columnIndex
	 * @param value
	 */
	public void setMat(int rowIndex, int columnIndex, double value);

	/**
	 * Set the objectives value of the P
	 *
	 * @param objective
	 *            of the LP
	 * @param target
	 *            type of objectives
	 * @return success indication
	 */
	public boolean setObjective(double[] objective, ObjectiveTargetType target);

	/**
	 * Set the objectives value of the LP
	 *
	 * @param objective
	 *            Objective of the ILP in terms of column id's
	 * @param target
	 *            type of objectives
	 * @return success indication
	 */
	public boolean setObjective(Map<Integer, Double> objective, ObjectiveTargetType target);

	/**
	 * Set the type of the objectives, either minimize or maximize.
	 *
	 * @param type
	 *            of objectives
	 * @return success indication
	 */
	public boolean setObjectiveType(ObjectiveTargetType type);

	public boolean setRightHandSide(int row, double rhs);

	public void setTimeOut(long seconds);

	/**
	 * Set a variable's type to be of type int.
	 *
	 * @param column
	 *            index of variable
	 * @return operation success indicaton
	 */
	public boolean setType(int column, VariableType type);

	/**
	 * Set a variable upperbound.
	 *
	 * @param column
	 *            identifier of variable
	 * @param upperBound
	 *            to assing to variable
	 * @return engineType manipulation succeeded
	 */
	public boolean setUpperBound(int column, double upperBound);

	/**
	 * Solve the LP and return the best variable assignment
	 *
	 * @return A non-empty map with variable-assignment pairs. An empty map if
	 *         no solution is found
	 */
	public Map<Integer, Double> solve();

	/**
	 * Solve the LP and return the best variable assignment and the assignment
	 * value.
	 *
	 * @return a non-empty map with assignment and associated value. Empty map
	 *         and -1 if no solution.
	 */
	public Pair<Map<Integer, Double>, Double> solveAndValue();

	/**
	 * Solve the LP and return the best variable assignment and the assignment
	 * value.
	 *
	 * @return a non-empty map with assignment and associated value. Empty map
	 *         and -1 if no solution.
	 */
	public Pair<double[], Double> solveAndValueNative();

	/**
	 * Solve the LP and return the best variable assignment
	 *
	 * @return
	 */
	public double[] solveNative();
	
	Pair<double[], Double> solveAndResetAndValueNative();

	public Pair<Pair<double[], Double>, Long> solveValueNativeSolveTimeNano();

	public void writeToFile(String dest);
}
